/*
 * Copyright (c) 2013, 2014, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.oracle.truffle.r.nodes.unary;

import com.oracle.truffle.api.*;
import com.oracle.truffle.api.dsl.*;
import com.oracle.truffle.api.frame.*;
import com.oracle.truffle.api.source.*;
import com.oracle.truffle.api.utilities.*;
import com.oracle.truffle.r.nodes.*;
import com.oracle.truffle.r.nodes.unary.UnaryArithmeticReduceNodeFactory.MultiElemStringHandlerFactory;
import com.oracle.truffle.r.runtime.*;
import com.oracle.truffle.r.runtime.data.*;
import com.oracle.truffle.r.runtime.ops.*;
import com.oracle.truffle.r.runtime.ops.na.*;

@NodeChild(value = "naRm", type = RNode.class)
public abstract class UnaryArithmeticReduceNode extends UnaryNode {

    @Child private MultiElemStringHandler stringHandler;

    private final BinaryArithmeticFactory factory;

    @Child private BinaryArithmetic arithmetic;

    private final ReduceSemantics semantics;

    private final NACheck na = NACheck.create();

    final ConditionProfile naRmProfile = ConditionProfile.createBinaryProfile();

    public UnaryArithmeticReduceNode(ReduceSemantics semantics, BinaryArithmeticFactory factory) {
        this.factory = factory;
        this.semantics = semantics;
        this.arithmetic = factory.create();
    }

    public UnaryArithmeticReduceNode(UnaryArithmeticReduceNode op) {
        // we recreate the arithmetic each time this specialization specializes
        // it also makes sense for polymorphic variations of this node
        this(op.semantics, op.factory);
    }

    private String handleString(VirtualFrame frame, RStringVector operand, byte naRm, int offset) {
        if (stringHandler == null) {
            CompilerDirectives.transferToInterpreterAndInvalidate();
            stringHandler = insert(MultiElemStringHandlerFactory.create(semantics, factory, na, null, null, null));
        }
        return stringHandler.executeString(frame, operand, naRm, offset);
    }

    protected boolean isNullInt() {
        return semantics.isNullInt();
    }

    @SuppressWarnings("unused")
    @Specialization(guards = "isNullInt")
    protected int doInt(RNull operand, byte naRm) {
        if (semantics.getEmptyWarning() != null) {
            RError.warning(semantics.emptyWarning);
        }
        return semantics.getIntStart();
    }

    @SuppressWarnings("unused")
    @Specialization(guards = "!isNullInt")
    protected double doDouble(RNull operand, byte naRm) {
        if (semantics.getEmptyWarning() != null) {
            RError.warning(semantics.emptyWarning);
        }
        return semantics.getDoubleStart();
    }

    @Specialization
    protected int doInt(int operand, byte naRm) {
        boolean profiledNaRm = naRmProfile.profile(naRm == RRuntime.LOGICAL_TRUE);
        na.enable(operand);
        if (profiledNaRm) {
            if (na.check(operand)) {
                if (semantics.getEmptyWarning() != null) {
                    RError.warning(semantics.emptyWarning);
                }
                return semantics.getIntStart();
            } else {
                return operand;
            }
        } else {
            return na.check(operand) ? RRuntime.INT_NA : operand;
        }
    }

    @Specialization
    protected double doDouble(double operand, byte naRm) {
        boolean profiledNaRm = naRmProfile.profile(naRm == RRuntime.LOGICAL_TRUE);
        na.enable(operand);
        if (profiledNaRm) {
            if (na.check(operand)) {
                if (semantics.getEmptyWarning() != null) {
                    RError.warning(semantics.emptyWarning);
                }
                return semantics.getIntStart();
            } else {
                return operand;
            }
        } else {
            return na.check(operand) ? RRuntime.DOUBLE_NA : operand;
        }
    }

    @Specialization
    protected int doLogical(byte operand, byte naRm) {
        boolean profiledNaRm = naRmProfile.profile(naRm == RRuntime.LOGICAL_TRUE);
        na.enable(operand);
        if (profiledNaRm) {
            if (na.check(operand)) {
                if (semantics.getEmptyWarning() != null) {
                    RError.warning(semantics.emptyWarning);
                }
                return semantics.getIntStart();
            } else {
                return operand;
            }
        } else {
            return na.check(operand) ? RRuntime.INT_NA : operand;
        }
    }

    @Specialization
    protected RComplex doComplex(RComplex operand, byte naRm) {
        if (semantics.supportComplex) {
            boolean profiledNaRm = naRmProfile.profile(naRm == RRuntime.LOGICAL_TRUE);
            na.enable(operand);
            if (profiledNaRm) {
                if (na.check(operand)) {
                    if (semantics.getEmptyWarning() != null) {
                        RError.warning(semantics.emptyWarning);
                    }
                    return RRuntime.double2complex(semantics.getDoubleStart());
                } else {
                    return operand;
                }
            } else {
                return na.check(operand) ? RRuntime.createComplexNA() : operand;
            }
        } else {
            throw RError.error(getEncapsulatingSourceSection(), RError.Message.INVALID_TYPE_ARGUMENT, "complex");
        }
    }

    @Specialization
    protected String doString(String operand, byte naRm) {
        if (semantics.supportString) {
            boolean profiledNaRm = naRmProfile.profile(naRm == RRuntime.LOGICAL_TRUE);
            na.enable(operand);
            if (profiledNaRm) {
                if (na.check(operand)) {
                    if (semantics.getEmptyWarning() != null) {
                        RError.warning(semantics.emptyWarning);
                    }
                    return semantics.getStringStart();
                } else {
                    return operand;
                }
            } else {
                return na.check(operand) ? RRuntime.STRING_NA : operand;
            }
        } else {
            throw RError.error(getEncapsulatingSourceSection(), RError.Message.INVALID_TYPE_ARGUMENT, "character");
        }
    }

    @SuppressWarnings("unused")
    @Specialization
    protected RRaw doString(RRaw operand, byte naRm) {
        throw RError.error(getEncapsulatingSourceSection(), RError.Message.INVALID_TYPE_ARGUMENT, "raw");
    }

    @Specialization
    protected int doIntVector(RIntVector operand, byte naRm) {
        boolean profiledNaRm = naRmProfile.profile(naRm == RRuntime.LOGICAL_TRUE);
        int result = semantics.getIntStart();
        na.enable(operand);
        int opCount = 0;
        for (int i = 0; i < operand.getLength(); i++) {
            int d = operand.getDataAt(i);
            na.enable(d);
            if (na.check(d)) {
                if (profiledNaRm) {
                    continue;
                } else {
                    return RRuntime.INT_NA;
                }
            } else {
                result = arithmetic.op(result, d);
            }
            opCount++;
        }
        if (opCount == 0 && semantics.getEmptyWarning() != null) {
            RError.warning(semantics.emptyWarning);
        }
        return result;
    }

    @Specialization
    protected double doDoubleVector(RDoubleVector operand, byte naRm) {
        boolean profiledNaRm = naRmProfile.profile(naRm == RRuntime.LOGICAL_TRUE);
        double result = semantics.getDoubleStart();
        na.enable(operand);
        int opCount = 0;
        for (int i = 0; i < operand.getLength(); i++) {
            double d = operand.getDataAt(i);
            na.enable(d);
            if (na.check(d)) {
                if (profiledNaRm) {
                    continue;
                } else {
                    return RRuntime.DOUBLE_NA;
                }
            } else {
                result = arithmetic.op(result, d);
            }
            opCount++;
        }
        if (opCount == 0 && semantics.getEmptyWarning() != null) {
            RError.warning(semantics.emptyWarning);
        }
        return result;
    }

    @Specialization
    protected int doLogicalVector(RLogicalVector operand, byte naRm) {
        boolean profiledNaRm = naRmProfile.profile(naRm == RRuntime.LOGICAL_TRUE);
        int result = semantics.getIntStart();
        na.enable(operand);
        int opCount = 0;
        for (int i = 0; i < operand.getLength(); i++) {
            byte d = operand.getDataAt(i);
            na.enable(d);
            if (na.check(d)) {
                if (profiledNaRm) {
                    continue;
                } else {
                    return RRuntime.INT_NA;
                }
            } else {
                result = arithmetic.op(result, d);
            }
            opCount++;
        }
        if (opCount == 0 && semantics.getEmptyWarning() != null) {
            RError.warning(semantics.emptyWarning);
        }
        return result;
    }

    @Specialization
    protected int doIntSequence(RIntSequence operand, @SuppressWarnings("unused") byte naRm) {
        int result = semantics.getIntStart();
        int current = operand.getStart();
        int opCount = 0;
        for (int i = 0; i < operand.getLength(); ++i) {
            result = arithmetic.op(result, current);
            current += operand.getStride();
        }
        if (opCount == 0 && semantics.getEmptyWarning() != null) {
            RError.warning(semantics.emptyWarning);
        }
        return result;
    }

    @Specialization
    protected double doDoubleSequence(RDoubleSequence operand, @SuppressWarnings("unused") byte naRm) {
        double result = semantics.getDoubleStart();
        double current = operand.getStart();
        int opCount = 0;
        for (int i = 0; i < operand.getLength(); ++i) {
            result = arithmetic.op(result, current);
            current += operand.getStride();
        }
        if (opCount == 0 && semantics.getEmptyWarning() != null) {
            RError.warning(semantics.emptyWarning);
        }
        return result;
    }

    @Specialization
    protected RComplex doComplexVector(RComplexVector operand, byte naRm) {
        if (semantics.supportComplex) {
            boolean profiledNaRm = naRmProfile.profile(naRm == RRuntime.LOGICAL_TRUE);
            RComplex result = RRuntime.double2complex(semantics.getDoubleStart());
            int opCount = 0;
            for (int i = 0; i < operand.getLength(); ++i) {
                RComplex current = operand.getDataAt(i);
                na.enable(current);
                if (na.check(current)) {
                    if (profiledNaRm) {
                        continue;
                    } else {
                        return RRuntime.createComplexNA();
                    }
                } else {
                    result = arithmetic.op(result.getRealPart(), result.getImaginaryPart(), current.getRealPart(), current.getImaginaryPart());
                }
                opCount++;
            }
            if (opCount == 0 && semantics.getEmptyWarning() != null) {
                RError.warning(semantics.emptyWarning);
            }
            return result;
        } else {
            throw RError.error(getEncapsulatingSourceSection(), RError.Message.INVALID_TYPE_ARGUMENT, "complex");

        }
    }

    // the algorithm that works for other types (reducing a vector starting with the "start value")
    // does not work for String-s as, in particular, we cannot supply the (lexicographically)
    // "largest" String for the implementation of max function

    @SuppressWarnings("unused")
    protected static String doStringVectorEmptyInternal(RStringVector operand, byte naRm, ReduceSemantics semantics, SourceSection sourceSection) {
        if (semantics.supportString) {
            if (semantics.getEmptyWarning() != null) {
                RError.warning(semantics.emptyWarning);
            }
            return semantics.getStringStart();
        } else {
            throw RError.error(sourceSection, RError.Message.INVALID_TYPE_ARGUMENT, "character");
        }
    }

    @Specialization(guards = "empty")
    protected String doStringVectorEmpty(RStringVector operand, byte naRm) {
        return doStringVectorEmptyInternal(operand, naRm, semantics, getEncapsulatingSourceSection());
    }

    @Specialization(guards = "lengthOne")
    protected String doStringVectorOneElem(RStringVector operand, byte naRm) {
        if (semantics.supportString) {
            boolean profiledNaRm = naRmProfile.profile(naRm == RRuntime.LOGICAL_TRUE);
            String result = operand.getDataAt(0);
            if (profiledNaRm) {
                na.enable(result);
                if (na.check(result)) {
                    return doStringVectorEmpty(operand, naRm);
                }
            }
            return result;
        } else {
            throw RError.error(getEncapsulatingSourceSection(), RError.Message.INVALID_TYPE_ARGUMENT, "character");
        }
    }

    @Specialization(guards = "longerThanOne")
    protected String doStringVector(VirtualFrame frame, RStringVector operand, byte naRm) {
        if (semantics.supportString) {
            return handleString(frame, operand, naRm, 0);
        } else {
            throw RError.error(getEncapsulatingSourceSection(), RError.Message.INVALID_TYPE_ARGUMENT, "character");
        }
    }

    @SuppressWarnings("unused")
    @Specialization
    protected RRaw doString(RRawVector operand, byte naRm) {
        throw RError.error(getEncapsulatingSourceSection(), RError.Message.INVALID_TYPE_ARGUMENT, "raw");
    }

    protected boolean empty(RStringVector vector) {
        return vector.getLength() == 0;
    }

    protected boolean lengthOne(RStringVector vector) {
        return vector.getLength() == 1;
    }

    protected boolean longerThanOne(RStringVector vector) {
        return vector.getLength() > 1;
    }

    public static final class ReduceSemantics {

        private final int intStart;
        private final double doubleStart;
        private final String stringStart = RRuntime.STRING_NA; // does not seem to change
        private final boolean nullInt;
        private final RError.Message emptyWarning;
        private final boolean supportComplex;
        private final boolean supportString;

        public ReduceSemantics(int intStart, double doubleStart, boolean nullInt, RError.Message emptyWarning, boolean supportComplex, boolean supportString) {
            this.intStart = intStart;
            this.doubleStart = doubleStart;
            this.nullInt = nullInt;
            this.emptyWarning = emptyWarning;
            this.supportComplex = supportComplex;
            this.supportString = supportString;
        }

        public int getIntStart() {
            return intStart;
        }

        public double getDoubleStart() {
            return doubleStart;
        }

        public String getStringStart() {
            return stringStart;
        }

        public boolean isNullInt() {
            return nullInt;
        }

        public RError.Message getEmptyWarning() {
            return emptyWarning;
        }

        public boolean supportsComplex() {
            return supportComplex;
        }

        public boolean supportsString() {
            return supportString;
        }

    }

    @NodeChildren({@NodeChild("operand"), @NodeChild("naRm"), @NodeChild("offset")})
    protected abstract static class MultiElemStringHandler extends RNode {

        public abstract String executeString(VirtualFrame frame, RStringVector operand, byte naRm, int offset);

        @Child private MultiElemStringHandler recursiveStringHandler;
        private final ReduceSemantics semantics;
        private final BinaryArithmeticFactory factory;
        @Child private BinaryArithmetic arithmetic;
        private final NACheck na;
        final ConditionProfile naRmProfile = ConditionProfile.createBinaryProfile();

        public MultiElemStringHandler(ReduceSemantics semantics, BinaryArithmeticFactory factory, NACheck na) {
            this.semantics = semantics;
            this.factory = factory;
            this.arithmetic = factory.create();
            this.na = na;
        }

        public MultiElemStringHandler(MultiElemStringHandler other) {
            this(other.semantics, other.factory, other.na);
        }

        private String handleString(VirtualFrame frame, RStringVector operand, byte naRm, int offset) {
            if (recursiveStringHandler == null) {
                CompilerDirectives.transferToInterpreterAndInvalidate();
                recursiveStringHandler = insert(MultiElemStringHandlerFactory.create(semantics, factory, na, null, null, null));
            }
            return recursiveStringHandler.executeString(frame, operand, naRm, offset);
        }

        @Specialization
        protected String doStringVectorMultiElem(VirtualFrame frame, RStringVector operand, byte naRm, int offset) {
            boolean profiledNaRm = naRmProfile.profile(naRm == RRuntime.LOGICAL_TRUE);
            String result = operand.getDataAt(offset);
            na.enable(result);
            if (profiledNaRm) {
                if (na.check(result)) {
                    // the following is meant to eliminate leading NA-s
                    if (offset == operand.getLength() - 1) {
                        // last element - all other are NAs
                        return doStringVectorEmptyInternal(operand, naRm, semantics, getEncapsulatingSourceSection());
                    } else {
                        return handleString(frame, operand, naRm, offset + 1);
                    }
                }
            } else {
                if (na.check(result)) {
                    return result;
                }
            }
            // when we reach here, it means that we have already seen one non-NA element
            assert !RRuntime.isNA(result);
            for (int i = offset + 1; i < operand.getLength(); ++i) {
                String current = operand.getDataAt(i);
                na.enable(current);
                if (na.check(current)) {
                    if (profiledNaRm) {
                        // skip NA-s
                        continue;
                    } else {
                        return RRuntime.STRING_NA;
                    }
                } else {
                    result = arithmetic.op(result, current);
                }
            }
            return result;
        }

    }

}
